local tABT = {}

-- Some predefined values
local anchorsConfig = {
	width = 200,
	height = 100,
	title = "TwinABT: Buffs",
	title2 = "TwinABT: Debuffs"
}
context = context or UI.CreateContext('TwinABTContext')
playerID = playerID or Inspect.Unit.Lookup('player')
local playerAvailabe = false
local refreshRate = 0.5
local refreshNow
--

function init()
	table.insert(Event.Addon.SavedVariables.Load.End, {function(identifier) if identifier == "TwinABT" then tABT:manageSavedVariables() end end, "TwinABT", "Saved variables loaded"})
end

-- Returns a proper time format and suffix for buff remaining time
local function getTimeString(remaining)
	local rTime, rSuffix
	if remaining <= 60 then
		rTime = math.floor(remaining)
		rSuffix = 's'
	elseif remaining <= 3600 then
		rTime = math.floor(remaining / 60)
		rSuffix = 'm'
	else
		rTime = math.floor(remaining / 3600)
		rSuffix = 'h'
	end
	
	return rTime, rSuffix
end

-- Checks if mouse is inside given frame
local function MouseInFrame(frame)
	local mouse = Inspect.Mouse()
	local frameX1, frameX2 = frame:GetLeft(), frame:GetLeft()+frame:GetWidth()
	local frameY1, frameY2 = frame:GetTop(), frame:GetTop()+frame:GetHeight()
	
	return (frameX1-mouse.x <= 0) and (frameX2-mouse.x >= 0) and (frameY1-mouse.y <= 0) and (frameY2-mouse.y >= 0)
end

function enableEditMode(enable)

	if enable then
		-- Unlocking anchor movement
		tABTbuffs.locked = false
		-- Hiding contents
		tABTbuffs.holders.buffs["player"]:SetVisible(false)
		tABTbuffs.holders.buffs["group"]:SetVisible(false)
		tABTbuffs.holders.buffs["debuff"]:SetVisible(false)
		-- enabling EDIT mode
		tABTbuffs.anchors.buffs:SetVisible(true)
		tABTbuffs.anchors.debuffs:SetVisible(true)
	else
		-- Locking anchor movement
		tABTbuffs.locked = true
		-- Showing contents
		tABTbuffs.holders.buffs["player"]:SetVisible(true)
		tABTbuffs.holders.buffs["group"]:SetVisible(true)
		tABTbuffs.holders.buffs["debuff"]:SetVisible(true)
		-- disabling EDIT mode
		tABTbuffs.anchors.buffs:SetVisible(false)
		tABTbuffs.anchors.debuffs:SetVisible(false)
		-- Refreshing buffs
		tABT:RefreshBuffsList()
	end
end

-- Creates and returns a frame for anchoring player's buffs
function tABT:GetBuffsAnchor()
	local anchor = UI.CreateFrame("Frame", "TwinABT Buffs Anchor", context)
	if not tABToptions.anchors.buffs.x then
		tABToptions.anchors.buffs.x = (UIParent:GetWidth() - anchorsConfig.width) / 2
		tABToptions.anchors.buffs.y = (UIParent:GetHeight() - anchorsConfig.height) / 2
	end
	anchor:SetPoint("TOPLEFT", UIParent, "TOPLEFT", tABToptions.anchors.buffs.x, tABToptions.anchors.buffs.y)
	anchor:SetWidth(anchorsConfig.width)
	anchor:SetHeight(anchorsConfig.height)
	anchor:SetBackgroundColor(0,0,0,0.5)
	anchor:SetVisible(false)
	-- Title
	anchor.title = UI.CreateFrame("Text", "TwinABT Buffs Anchor Title", anchor)
	anchor.title:SetPoint("CENTER", anchor, "CENTER", 0, 0)
	anchor.title:SetText(anchorsConfig.title)
	anchor.title:SetFontColor(1,1,1)
	anchor.title:SetFontSize(14)
	
	-- Dragable
	function anchor.Event:LeftDown()
		self.drag = true
	end
	function anchor.Event:LeftUp()
		self.drag = false
		self.move = false
		self.offset = nil
		tABToptions.anchors.buffs.x = self:GetLeft()
		tABToptions.anchors.buffs.y = self:GetTop()
	end
	function anchor.Event:MouseMove(x,y)
		if self.drag then
			if not self.offset then
				self.offset = {
					x = x-self:GetLeft(),
					y = y-self:GetTop()
				}
			end
			local moveX, moveY = x-self.offset.x, y-self.offset.y
			anchor:SetPoint("TOPLEFT", UIParent, "TOPLEFT", moveX, moveY)
		end
	end
	
	return anchor
end

-- Creates and returns a frame for anchoring player's debuffs
function tABT:GetDebuffsAnchor()
	local anchor = UI.CreateFrame("Frame", "TwinABT Debuffs Anchor", context)
	if not tABToptions.anchors.debuffs.x then
		tABToptions.anchors.debuffs.x = (UIParent:GetWidth() - anchorsConfig.width) / 2
		tABToptions.anchors.debuffs.y = (UIParent:GetHeight() - anchorsConfig.height) / 2
	end
	anchor:SetPoint("TOPLEFT", UIParent, "TOPLEFT", tABToptions.anchors.debuffs.x, tABToptions.anchors.debuffs.y)
	anchor:SetWidth(anchorsConfig.width)
	anchor:SetHeight(anchorsConfig.height / 2)
	anchor:SetBackgroundColor(0,0,0,0.5)
	anchor:SetVisible(false)
	-- Title
	anchor.title = UI.CreateFrame("Text", "TwinABT Debuffs Anchor Title", anchor)
	anchor.title:SetPoint("CENTER", anchor, "CENTER", 0, 0)
	anchor.title:SetText(anchorsConfig.title2)
	anchor.title:SetFontColor(1,1,1)
	anchor.title:SetFontSize(14)
	anchor.title:SetAlpha(1)
	
	-- Dragable
	function anchor.Event:LeftDown()
		self.drag = true
	end
	function anchor.Event:LeftUp()
		self.drag = false
		self.move = false
		self.offset = nil
		tABToptions.anchors.debuffs.x = self:GetLeft()
		tABToptions.anchors.debuffs.y = self:GetTop()
	end
	function anchor.Event:MouseMove(x,y)
		if self.drag then
			if not self.offset then
				self.offset = {
					x = x-self:GetLeft(),
					y = y-self:GetTop()
				}
			end
			local moveX, moveY = x-self.offset.x, y-self.offset.y
			anchor:SetPoint("TOPLEFT", UIParent, "TOPLEFT", moveX, moveY)
		end
	end
	
	return anchor
end

-- Creates and returns a holder frame
function tABT:GetHolder(anchor, anchor2 ,offsetY)
	local holder = UI.CreateFrame("Frame", "TwinABT Buffs Holder", context)
	if anchor then
		holder:SetPoint("TOPRIGHT", anchor, "TOPRIGHT", 0, offsetY)
	else
		holder:SetPoint("TOPRIGHT", anchor2, "BOTTOMRIGHT", 0, offsetY)
	end
	holder:SetWidth(1)
	holder:SetHeight(50)
	holder:SetVisible(true)
	
	return holder
end

-- Creates and returns content of holder
function tABT:GetContent(anchor, parent, position)
	local content = UI.CreateFrame("Frame", "TwinABT Buffs Holder Content", parent)
	local parentPoint = (position == 1 and "TOPRIGHT") or "TOPLEFT"
	content:SetPoint("TOPRIGHT", anchor, parentPoint, -1, 0)
	content:SetHeight(50)
	
	return content
end

-- Returns a first occurence of buff without duration
local function getNextBuffWithoutDuration(list, buffs, lastRemove)
	lastRemove = lastRemove or 1
	-- Clearing list from buffs without duration
	for i=lastRemove, #list do
		local id = list[i]
		if not buffs[id].duration then return i, id end
	end
	return nil
end
 
-- Sorts list
-- Sort order (right to left): auras (no duration) -> buff duration (from longest to smallest) 
function tABT:SortBuffList(buffGroup)
	local list = tABTbuffs.buffList.sorted[buffGroup]
	local buffs = tABTbuffs.buffList[buffGroup]
	local newList = {}
	
	-- Getting all items without duration
	local pos = 1
	while pos do
		pos, id = getNextBuffWithoutDuration(list, buffs, pos)
		if pos then
			table.insert(newList, id)
			table.remove(list, pos)
		end
	end
	
	-- Sorting list by duration
	table.sort(list, function(a, b) 
		if not buffs[a] or not buffs[b] then return false 
		else return buffs[a].remaining > buffs[b].remaining end
	end)
	
	-- Retrieving sorted data
	for i,v in ipairs(list) do
		table.insert(newList, v)
	end
	
	-- Setting newly sorted list
	tABTbuffs.buffList.sorted[buffGroup] = newList
end


-- makes a full refresh of current buffs on the player
function tABT:RefreshBuffsList()
	-- Getting data about player's buffs
	local buffList = Inspect.Buff.List('player')
	local buffsData = Inspect.Buff.Detail('player', buffList)
	-- Resetting lists
	tABTbuffs.buffList = {["player"] = {}, ["group"] = {}, ["debuff"] = {}}
	tABTbuffs.buffList.sorted = {["player"] = {}, ["group"] = {}, ["debuff"] = {}}
	tABTgroups.ResetGroups()
	-- Sorting the list to: player/group/debuff
	if tABTbuffs.tempBuffs then tABTbuffs.tempBuffs = nil end
	for buffID,buff in pairs(buffsData) do
		if buff.debuff then
			tABTbuffs.buffList.debuff[buffID] = buff
			table.insert(tABTbuffs.buffList.sorted.debuff, buffID)
		else
			local caster = (buff.caster == playerID and "player") or "group"
			tABTbuffs.buffList[caster][buffID] = buff
		end
	end
	-- Applying grouping
	for _,group in ipairs(tABTgroups) do
		tABTgroups.RetrieveAll(group, tABTbuffs.buffList["player"], tABTbuffs.buffList["group"])
	end
	-- Managing leftover buffs
	for id,_ in pairs(tABTbuffs.buffList["player"]) do
		table.insert(tABTbuffs.buffList.sorted["player"], id)
	end
	for id,_ in pairs(tABTbuffs.buffList["group"]) do
		table.insert(tABTbuffs.buffList.sorted["group"], id)
	end
	
	-- Sorting the new buff lists
	tABT:SortBuffList("player")
	tABT:SortBuffList("group")
	tABT:SortBuffList("debuff")
	-- Displaying the new lists
	
	tABTgroups.UpdateGroupsContent("player", tABTbuffs.holders.buffs["player"].content.buffGroups)
	tABTgroups.UpdateGroupsContent("group", tABTbuffs.holders.buffs["group"].content.buffGroups)
	
	tABTbuffmanager.DrawBuffs("player", true)
	tABTbuffmanager.DrawBuffs("group", true)
	tABTbuffmanager.DrawBuffs("debuff", true)
	
end

-- adds the new buff
function tABT:BuffAdd(buffs)
	-- getting buffs details
	buffData = Inspect.Buff.Detail(playerID, buffs)
	local groupsChanged = {}
	-- working with each buff
	for buffID,buff in pairs(buffData) do
		if buff.debuff then
			tABTbuffs.buffList.debuff[buffID] = buff
			tABTbuffmanager.addBuffToList(tABTbuffs.buffList.sorted["debuff"], tABTbuffs.buffList["debuff"], buff)
			groupsChanged["debuff"] = true
		else
			local buffGroup = (buff.caster == playerID and "player") or ("group")
			-- deciding if buff needs to be grouped
			local isGrouped = false
			for _,group in ipairs(tABTgroups) do
				if not isGrouped and tABTgroups.CheckBuff(group, buff, buffGroup) then
					isGrouped = true
					groupsChanged[buffGroup] = true
					break
				end
			end
			-- Adding buff to buff list
			if not isGrouped then
				tABTbuffs.buffList[buffGroup][buffID] = buff
				tABTbuffmanager.addBuffToList(tABTbuffs.buffList.sorted[buffGroup], tABTbuffs.buffList[buffGroup], buff)
				groupsChanged[buffGroup] = true
			end
		end
	end
	
	-- Redrawing changed buff list
	for buffGroup,_ in pairs(groupsChanged) do
		tABTbuffmanager.DrawBuffs(buffGroup)
		if buffGroup ~= "debuff" then
			tABTgroups.UpdateGroupsContent(buffGroup, tABTbuffs.holders.buffs[buffGroup].content.buffGroups)
		end
	end	

	-- Redrawing group details if needed
	if tABTgroups.groupDetails:GetVisible() then
		tABTgroups.groupDetails.update()
	end
end

-- Removes the buff from the list
function tABT:BuffRemove(buffs)
	local groupsChanged = {}
	local buffsChanged = {}
	for buffID,_ in pairs(buffs) do 
		-- Hiding buff details frame
		if tABTbuffmanager.buffDetails.buff == buffID then
			tABTbuffmanager.buffDetails:SetVisible(false)
			tABTbuffmanager.buffDetails.buff = nil
		end
		-- Checking if buff was grouped
		local grouped = tABTgroups.RemoveBuff(buffID)
		if grouped then
			groupsChanged[grouped] = true
		end
		
		-- working with ungrouped buff
		if not grouped then
			local buffGroup = (tABTbuffs.buffList["debuff"][buffID] and "debuff") or (tABTbuffs.buffList["player"][buffID] and "player") or ("group")
			-- Removing buff details from memory
			tABTbuffs.buffList[buffGroup][buffID] = nil
			-- Removing buff from list to display
			for i,id in ipairs(tABTbuffs.buffList.sorted[buffGroup]) do
				if buffID == id then table.remove(tABTbuffs.buffList.sorted[buffGroup], i) break end
			end
			buffsChanged[buffGroup] = true
		end
		
	end
	-- Redrawing changed buff list
	for buffGroup,_ in pairs(buffsChanged) do
		tABTbuffmanager.DrawBuffs(buffGroup)
	end	
	-- Redrawing changed group list
	for buffGroup,_ in pairs(groupsChanged) do
		tABTgroups.UpdateGroupsContent(buffGroup, tABTbuffs.holders.buffs[buffGroup].content.buffGroups)
	end	
end

-- Called when players buffs status changed (using for catching a change in stacks)
function tABT:BuffChange(buffs)
	-- Gettings buffs data
	local buffData = Inspect.Buff.Detail(playerID, buffs)
	for buffID,buff in pairs(buffData) do
		-- stored buff data
		local buffGroup = (buff.debuff and "debuff") or (buff.caster == playerID and "player") or ("group")
		local savedBuff = tABTbuffs.buffList[buffGroup][buffID]
		-- Checking if buff has stacks
		if savedBuff and (buff.stack or savedBuff.stack) then
			-- Getting proper lists
			local boxes = tABTbuffs.holders.buffs[buffGroup].content.buffs.boxes
			local list = tABTbuffs.buffList.sorted[buffGroup]
			-- Finding the box of the buff
			local box
			for i,id in ipairs(list) do
				if id == buff.id then box = boxes[i] break end
			end
			-- Updating stack
			if buff.stack then
				box.stacks:SetText('' .. buff.stack)
			else
				box.stacks:SetAlpha(0)
				box.stacksBG:SetAlpha(0)
			end
			-- Updating stored buff data
			tABTbuffs.buffList[buffGroup][buffID] = buff
		end
	end

	-- Redrawing group details if needed
	if tABTgroups.groupDetails:GetVisible() then
			tABTgroups.groupDetails.update()
	end
end

-- Updates remaining time of active buffs
local function updateTime(buffGroup)
	-- Getting data
	local list = tABTbuffs.buffList.sorted[buffGroup]
	local buffs = tABTbuffs.buffList[buffGroup]
	local boxes = tABTbuffs.holders.buffs[buffGroup].content.buffs.boxes
	
	-- Finding all buffs that has remaining time and refreshing it
	for i, buffID in ipairs(list) do
		if buffs[buffID].remaining then
			local buff = Inspect.Buff.Detail("player", buffID)
			if buff then
				buffs[buffID] = buff
				local rTime, rSuffix = getTimeString(buff.remaining)
				-- Setting text
				boxes[i].text:SetText(rTime .. rSuffix)
			end
		end
	end
end

-- Updates remaining time on buffs (executed based on refresh rate)
function tABT:UpdateRemainingTime()
	if tABTbuffs.locked and playerAvailabe and refreshNow() then
		-- checking on 'LongBuffs' group
		-- player
		local inBuffsPlayer = tABTgroups.CheckLongBuffs("player")
		if #inBuffsPlayer > 0 then 
			for _,buff in ipairs(inBuffsPlayer) do
				tABTbuffs.buffList["player"][buff.id] = buff
				tABTbuffmanager.addBuffToList(tABTbuffs.buffList.sorted["player"], tABTbuffs.buffList["player"], buff)
			end
			tABTgroups.UpdateGroupsContent("player", tABTbuffs.holders.buffs["player"].content.buffGroups)
			tABTbuffmanager.DrawBuffs("player")
		end
		-- group
		local inBuffsGroup = tABTgroups.CheckLongBuffs("group")
		if #inBuffsGroup > 0 then 
			for _,buff in ipairs(inBuffsGroup) do
				tABTbuffs.buffList["group"][buff.id] = buff
				tABTbuffmanager.addBuffToList(tABTbuffs.buffList.sorted["group"], tABTbuffs.buffList["group"], buff.id)
			end
			tABTgroups.UpdateGroupsContent("group", tABTbuffs.holders.buffs["group"].content.buffGroups)
			tABTbuffmanager.DrawBuffs("group")
		end
		-- updating time
		updateTime("player")
		updateTime("group")
		updateTime("debuff")
		-- Redrawing group details if needed
		if tABTgroups.groupDetails:GetVisible() and tABTgroups.groupDetails.hasTime then
			tABTgroups.groupDetails.updateTime()
		end
		
	
		-- hiding GroupDetails box if it has time-out
		if tABTgroups.groupDetails:GetVisible() and tABTgroups.groupDetails.focus then
			-- checking if mouse is not inside box
			if not MouseInFrame(tABTgroups.groupDetails) then
				local cTime = os.time()
				if os.difftime(cTime, tABTgroups.groupDetails.time) > 3 then
					tABTgroups.groupDetails:SetVisible(false)
					tABTgroups.groupDetails.focus = false
				end
			end
		end
	end
	
end

-- Checks if time between frames updates exceeded refresh rate
function tABT:CheckIfRefreshRequired()
	local oldTime = os.time()
	return function()
		local cTime = os.time()
		difference = os.difftime(cTime, oldTime)
		if difference >= refreshRate then oldTime = cTime return true end
		return false
	end
end

-- Handles commands sent by user
function SlashHandler(args)
	if string.len(args) == 0 then
		-- Slash commands list
		print('Twin: Advanced Buff Tracker in-game commands list:')
		print("/twinABT default -- resets addon's elements position to default")
		print('/twinABT edit -- toggles edit mode')
		print('/twinABT reset -- makes a full reset of addon settings. Requires /reloadui afterwards. Errors are expected using this functionality')
	else
		if args == 'default' then
			-- buffs anchor
			tABToptions.anchors.buffs.x = (UIParent:GetWidth() - anchorsConfig.width) / 2
			tABToptions.anchors.buffs.y = (UIParent:GetHeight() - anchorsConfig.height) / 2
			tABTbuffs.anchors.buffs:ClearPoint("TOPLEFT")
			tABTbuffs.anchors.buffs:SetPoint("TOPLEFT", UIParent, "TOPLEFT", tABToptions.anchors.buffs.x, tABToptions.anchors.buffs.y)
			-- debuffs anchor
			tABToptions.anchors.debuffs.x = (UIParent:GetWidth() - anchorsConfig.width) / 2
			tABToptions.anchors.debuffs.y = (UIParent:GetHeight() - anchorsConfig.height) / 2
			tABTbuffs.anchors.debuffs:ClearPoint("TOPLEFT")
			tABTbuffs.anchors.debuffs:SetPoint("TOPLEFT", UIParent, "TOPLEFT", tABToptions.anchors.debuffs.x, tABToptions.anchors.debuffs.y)
			-- options frame
			tABToptions.mainFrame.x = (UIParent:GetWidth() - tABToptions.mainFrame.width) / 2
			tABToptions.mainFrame.y = (UIParent:GetHeight() - tABToptions.mainFrame.height) / 2
			tABToptions.mainFrame.frame:ClearPoint("TOPLEFT")
			tABToptions.mainFrame.frame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", tABToptions.mainFrame.x, tABToptions.mainFrame.y)
			-- TwinABT button	
			tABToptions.toggleButton.x = (UIParent:GetWidth() - tABToptions.toggleButton.width) / 2
			tABToptions.toggleButton.y = (UIParent:GetHeight() - tABToptions.toggleButton.height) / 2
			tABToptions.button:ClearPoint("TOPLEFT")
			tABToptions.button:SetPoint("TOPLEFT", UIParent, "TOPLEFT", tABToptions.toggleButton.x, tABToptions.toggleButton.y)
			
		elseif args == 'edit' then
			local display = not tABToptions.mainFrame.frame:GetVisible()
			tABToptions.mainFrame.frame:SetVisible(display)
			-- Buffs anchors
			tABTbuffs.EnableEditMode(display)
		elseif args == 'reset' then
			tABToptions = tABToptionsDefaults
			tABTgroups = tABTgroupsDefaults
			tABTbuffmanager = {}
			print('please do /reloadui for changes to take effect')
		end
	end
end

-- Manages saved variables
function tABT:manageSavedVariables()
	tABTbuffs = {}
	-- Buff anchor lock
	tABTbuffs.locked = true
	tABTbuffs.EnableEditMode = enableEditMode
	
	-- Preparing for showing buffs:
	-- Creating anchors
	tABTbuffs.anchors = { ["buffs"] = {}, ["debuffs"] = {} }
	tABTbuffs.anchors.buffs = tABT:GetBuffsAnchor()
	tABTbuffs.anchors.debuffs = tABT:GetDebuffsAnchor()
	
	-- Creating Holders
	tABTbuffs.holders = { buffs = { ["player"] = {}, ["group"] = {}}, debuffs = {} }
	tABTbuffs.holders.buffs["player"] = tABT:GetHolder(tABTbuffs.anchors.buffs, nil, 0)
	tABTbuffs.holders.buffs["group"] = tABT:GetHolder(nil, tABTbuffs.holders.buffs["player"], 3)
	tABTbuffs.holders.buffs["debuff"] = tABT:GetHolder(tABTbuffs.anchors.debuffs, nil, 0)
	
	-- Managing Holders:
	-- Adding appropriate Inner Holders
	tABTbuffs.holders.buffs["player"].content = { buffGroups = {}, buffs = {} }
	tABTbuffs.holders.buffs["player"].content.buffGroups = tABT:GetContent(tABTbuffs.holders.buffs["player"], tABTbuffs.holders.buffs["player"], 1)
	tABTbuffs.holders.buffs["player"].content.buffs = tABT:GetContent(tABTbuffs.holders.buffs["player"].content.buffGroups, tABTbuffs.holders.buffs["player"], 2)
		
	tABTbuffs.holders.buffs["group"].content = { buffGroups = {}, buffs = {} }
	tABTbuffs.holders.buffs["group"].content.buffGroups = tABT:GetContent(tABTbuffs.holders.buffs["group"], tABTbuffs.holders.buffs["group"], 1)
	tABTbuffs.holders.buffs["group"].content.buffs = tABT:GetContent(tABTbuffs.holders.buffs["group"].content.buffGroups, tABTbuffs.holders.buffs["group"], 2)
	
	tABTbuffs.holders.buffs["debuff"].content = { buffs = {} }
	tABTbuffs.holders.buffs["debuff"].content.buffs = tABT:GetContent(tABTbuffs.holders.buffs["debuff"], tABTbuffs.holders.buffs["debuff"], 1)
	
	-- Adding Content
	tABTgroups.AddGroupsToContent(tABTbuffs.holders.buffs["player"].content.buffGroups, "player")
	tABTgroups.AddGroupsToContent(tABTbuffs.holders.buffs["group"].content.buffGroups, "group")
	
	-- Player buffs/debuffs
	tABTbuffs.buffList = {["player"] = {}, ["group"] = {}, ["debuff"] = {}}
	tABTbuffs.buffList.sorted = {["player"] = {}, ["group"] = {}, ["debuff"] = {}}
	
	-- Adding Buffs Content Handlers
	-- Availability change (refreshing full buffs list)
	table.insert(Event.Unit.Availability.None, { function (units) if units[playerID] then playerAvailabe = false end end, "TwinABT", "Player availability changed" })
	table.insert(Event.Unit.Availability.Partial, { function (units) if units[playerID] then playerAvailabe = false end end, "TwinABT", "Player availability changed" })
	table.insert(Event.Unit.Availability.Full, { function (units) if units[playerID] then playerAvailabe = true tABT:RefreshBuffsList() end end, "TwinABT", "Player availability changed" })
	-- Basic buff events
	table.insert(Event.Buff.Add, { function(unit,buffs) if unit == playerID then tABT:BuffAdd(buffs) end end , "TwinABT", "Buff Added to Player" })
	table.insert(Event.Buff.Remove, { function(unit,buffs) if unit == playerID then tABT:BuffRemove(buffs) end end , "TwinABT", "Buff Removed from Player" })
	table.insert(Event.Buff.Change, { function(unit,buffs) if unit == playerID then tABT:BuffChange(buffs) end end, "TwinABT", "Buff Removed from Player" })
	-- Regular updates (time)
	refreshNow = tABT:CheckIfRefreshRequired()
	table.insert(Event.System.Update.Begin, { function() tABT:UpdateRemainingTime() end , "TwinABT", "Game frame refresh" })
	-- Slash commands
	table.insert(Command.Slash.Register("twinABT"), {SlashHandler, "TwinABT", "Slash Command"})
end


init()