local tABT = {}
context = context or UI.CreateContext('TwinABTContext')
playerID = playerID or Inspect.Unit.Lookup('player')


-- Returns a proper time format and suffix for buff remaining time
local function getTimeString(remaining)
	local rTime, rSuffix
	if remaining <= 60 then
		rTime = math.floor(remaining)
		rSuffix = 's'
	elseif remaining <= 3600 then
		rTime = math.floor(remaining / 60)
		rSuffix = 'm'
	else
		rTime = math.floor(remaining / 3600)
		rSuffix = 'h'
	end
	
	return rTime, rSuffix
end

-- Displays all buffs in list
local function DrawBuffs(buffGroup, force)

	local buffList = tABTbuffs.buffList[buffGroup]
	local sortedList = tABTbuffs.buffList.sorted[buffGroup]
	local content = tABTbuffs.holders.buffs[buffGroup].content.buffs
	content.boxes = content.boxes or {}
	
	local count = 0
	for i,id in ipairs(sortedList) do
		buff = buffList[id]
		-- Checking if there are enough boxes. If not - adding new box
		if not content.boxes[i] then content.boxes[i] = tABT:CreateBuffBox(content, content.boxes, i) end
		-- Updating box to match buff
		if force or not content.boxes[i].buffID or content.boxes[i].buffID ~= buff.id then
			-- Reseting box
			tABT:ResetBox(content.boxes[i])
			
			-- Breakpoint (buffs in row limit reached)
			if tABTbuffmanager.limitBuffInRow and i~=1 and (i-1)%tABTbuffmanager.buffInRow == 0 then
				content.boxes[i]:ClearPoint("TOPRIGHT")
				content.boxes[i]:SetPoint("TOPRIGHT", content.boxes[1+(tABTbuffmanager.buffInRow*(((i-1)/tABTbuffmanager.buffInRow)-1))], "BOTTOMRIGHT", 0, 21)
				content.boxes[i].breakpoint = true
			end
			
			-- Debuff color
			if buff.debuff then
				local color = {0,0,0}
				if buff.disease then color = {250/255,229/255,36/255} --yellow
				elseif buff.poison then color = {59/255,172/255,2/255} --green
				elseif buff.curse then color = {159/255,4/255,242/255} --pink
				end
				content.boxes[i]:SetBackgroundColor(unpack(color))
			end
			-- Setting values
			content.boxes[i].texture:SetTexture('Rift', buff.icon)
			content.boxes[i]:SetVisible(true)
			content.boxes[i].buffID = buff.id
			-- Remaining time
			if buff.remaining then
				local rTime, rSuffix = getTimeString(buff.remaining)
				-- Setting text
				content.boxes[i].text:SetAlpha(1)
				content.boxes[i].text:SetText(rTime .. rSuffix)
				content.boxes[i].textBG:SetAlpha(1)
			end
			-- Stacks
			if buff.stack then
				content.boxes[i].stacks:SetText('' .. buff.stack)
				content.boxes[i].stacks:SetAlpha(1)
				content.boxes[i].stacksBG:SetAlpha(1)
				content.boxes[i].stacksBG:SetHeight(content.boxes[i].stacks:GetHeight())
				content.boxes[i].stacksBG:SetWidth(content.boxes[i].stacks:GetWidth())
			end
		end
		count = i
	end
	
	-- fixing holder height
	local holder = tABTbuffs.holders.buffs[buffGroup]
	if not tABTbuffmanager.limitBuffInRow then
		holder:SetHeight(50)
	else
		local layers = 1
		if count > tABTbuffmanager.buffInRow then
			layers = ((count-(count%tABTbuffmanager.buffInRow))/tABTbuffmanager.buffInRow)
			if count%tABTbuffmanager.buffInRow > 0 then layers = layers + 1 end
		end
		holder:SetHeight((layers*50)+(1*(layers-1)))
	end
	
	-- Reseting all unused boxes
	if count+1 <= #content.boxes then
		for i=count+1, #content.boxes do
			tABT:ResetBox(content.boxes[i])
		end
	end
end
-- Creates and returns a Buff Box
function tABT:CreateBuffBox(parent, list, index)
	local boxConfig = {
		width = 30,
		height = 30,
		spacing = 1,
		border = 1,
		text = {
			fontSize = 13,
			fontColor = {1,1,1,1},
			layer = 2,
			height = 20,
			background = {0, 0, 0, 0.8}
		},
		stacks = {
			fontSize = 12,
			fontColor = {1,1,1,1},
			layer = 2,
			background = {0,0,0,0.9}
		},
	}
	-- Border
	local buffBox = UI.CreateFrame('Frame', 'TwinABT Buffs Box', parent)
	local offset = (#list) * (boxConfig.width + boxConfig.spacing) * (-1)
	if list[index-1] then
		buffBox:SetPoint('TOPRIGHT', list[index-1], 'TOPLEFT', -1, 0)
		buffBox.infront = list[index-1]
	else
		buffBox:SetPoint('TOPRIGHT', parent, 'TOPRIGHT', 0, 0)
		buffBox.infront = nil
	end
	buffBox:SetWidth(boxConfig.width)
	buffBox:SetHeight(boxConfig.height)
	buffBox:SetBackgroundColor(0,0,0)
	buffBox:SetVisible(false)
	-- Texture
	local buffTexture = UI.CreateFrame('Texture', 'TwinABT Buffs Box', buffBox)
	buffTexture:SetPoint('CENTER', buffBox, 'CENTER', 0, 0)
	buffTexture:SetWidth(boxConfig.width - (boxConfig.border * 2))
	buffTexture:SetHeight(boxConfig.height - (boxConfig.border * 2))
	-- Text
	local boxText = UI.CreateFrame('Text', 'TwinABT Buffs Box Text', buffBox)
	boxText:SetPoint('CENTERTOP', buffBox, 'CENTERBOTTOM', 0, 1)
	boxText:SetLayer(boxConfig.text.layer)
	boxText:SetFontSize(boxConfig.text.fontSize)
	boxText:SetFontColor(unpack(boxConfig.text.fontColor))
	-- Text Background
	local boxTextBG = UI.CreateFrame('Frame', 'TwinABT Buffs Box Text Background', buffBox)
	boxTextBG:SetPoint('CENTERTOP', buffBox, 'CENTERBOTTOM', 0, 0)
	boxTextBG:SetLayer(boxConfig.text.layer-1)
	boxTextBG:SetHeight(boxConfig.text.height)
	boxTextBG:SetWidth(boxConfig.width)
	boxTextBG:SetBackgroundColor(unpack(boxConfig.text.background))
	-- Stacks
	local boxStacks = UI.CreateFrame('Text', 'TwinABT Buffs Box Stacks', buffBox)
	boxStacks:SetPoint('TOPRIGHT', buffBox, 'TOPRIGHT', 0, 0)
	boxStacks:SetLayer(boxConfig.stacks.layer)
	boxStacks:SetFontSize(boxConfig.stacks.fontSize)
	boxStacks:SetFontColor(unpack(boxConfig.stacks.fontColor))
	-- Stacks shadow
	local boxStacksBG = UI.CreateFrame('Frame', 'TwinABT Buffs Box Stacks Background', buffBox)
	boxStacksBG:SetPoint('CENTER', boxStacks, 'CENTER', 0, 0)
	boxStacksBG:SetLayer(boxConfig.stacks.layer - 1)
	boxStacksBG:SetBackgroundColor(unpack(boxConfig.stacks.background))
	
	-- Keeping track
	buffBox.texture = buffTexture
	buffBox.text = boxText
	buffBox.textBG = boxTextBG
	buffBox.stacks = boxStacks
	buffBox.stacksBG = boxStacksBG
	
	-- Mouse Over: Buff details
	function buffBox.Event:MouseIn()
		if self.buffID then MouseOver(self) end
	end
	
	-- Mouse Out: hide buff details
	function buffBox.Event:MouseOut()
		if buffDetails.buff == self.buffID then buffDetails:SetVisible(false) end
	end
	
	-- Mouse Right Click: remove buff
	function buffBox.Event:RightClick()
		if self.buffID then
			-- Hiding buff details
			if buffDetails.buff == self.buffID then buffDetails:SetVisible(true) end
			-- Canceling buff
			Command.Buff.Cancel(self.buffID) 
		end
	end
	
	-- Adding new box to list
	return buffBox
end

-- Displays Buff Details
function MouseOver(box)
	-- Getting latest buff details
	local buff = Inspect.Buff.Detail('player', box.buffID)
	
	-- if buff still exists (possible latency problems)
	if buff then
		if not buff.descriptionComplete then Command.Buff.Describe(playerID, box.buffID) end
		
		-- Showing Data
		-- Points
		tABTbuffmanager.buffDetails:ClearPoint('TOPRIGHT')
		tABTbuffmanager.buffDetails:SetPoint('TOPRIGHT', box, 'BOTTOMRIGHT', 0, 1)
		-- Texts: caster
		tABTbuffmanager.buffDetails.name:SetText(buff.name)
		local unit = Inspect.Unit.Detail(buff.caster)
		local caster = (not unit and buff.debuff and "Unknown") or (not unit and "Raid Buff") or (unit.id == playerID and "Self") or (unit.name)
		--if not caster then caster = 'Raid Buff' elseif caster.id == playerID then caster = 'Self' else caster = caster.name end
		tABTbuffmanager.buffDetails.caster:SetText('Caster: ' .. caster)
		-- Texts: description
		local runeDescription = "A rune is temporarily augmenting one of your items."
		tABTbuffmanager.buffDetails.description:SetText((not buff.description and '') or(buff.rune and runeDescription) or buff.description)
		-- Visuals
		tABTbuffmanager.buffDetails:SetHeight(tABTbuffmanager.buffDetails.name:GetHeight() + tABTbuffmanager.buffDetails.caster:GetHeight() + tABTbuffmanager.buffDetails.description:GetHeight() + 15)
		if tABTbuffmanager.buffDetails.name:GetWidth() > 240 then tABTbuffmanager.buffDetails:SetWidth(buffDetails.name:GetWidth() + 10) else tABTbuffmanager.buffDetails:SetWidth(250) end
		
		tABTbuffmanager.buffDetails:SetVisible(true)
		tABTbuffmanager.buffDetails.buff = buff.id
	end
end

-- Resets box to default values
function tABT:ResetBox(box)
	-- box
	box:SetVisible(false)
	box:SetBackgroundColor(0,0,0)
	box.buffID = nil
	if box.breakpoint then
		box:ClearPoint("TOPRIGHT")
		box:SetPoint('TOPRIGHT', box.infront, 'TOPLEFT', -1, 0)
	end
	box.breakpoint = false
	-- time
	box.text:SetAlpha(0)
	box.textBG:SetAlpha(0)
	-- stacks
	box.stacks:SetAlpha(0)
	box.stacksBG:SetAlpha(0)
end

-- Creates and returns a buff details frame
function tABT:GetBuffDetailsFrame()
	-- Buff Details
	buffDetails = UI.CreateFrame('Frame', 'TwinABTBuffDetails', context)
	buffDetails:SetWidth(250)
	buffDetails:SetHeight(50)
	buffDetails:SetLayer(5)
	buffDetails:SetBackgroundColor(0,0,0,0.9)
	buffDetails:SetVisible(false)
	-- Buff Details: Name
	buffDetails.name = UI.CreateFrame('Text', 'TwinABTBuffDetails.Name', buffDetails)
	buffDetails.name:SetPoint('TOPLEFT', buffDetails, 'TOPLEFT', 5, 5)
	buffDetails.name:SetWidth(240)
	buffDetails.name:SetFontColor(237/255,191/255,47/255)
	buffDetails.name:SetFontSize(15)
	-- Buff Details: Caster
	buffDetails.caster = UI.CreateFrame('Text', 'TwinABTBuffDetails.Caster', buffDetails)
	buffDetails.caster:SetPoint('TOPLEFT', buffDetails.name, 'BOTTOMLEFT', 0, 0)
	buffDetails.caster:SetWidth(240)
	buffDetails.caster:SetFontColor(195/255,185/255,139/255)
	buffDetails.caster:SetFontSize(13)
	-- Buff Details: Description
	buffDetails.description = UI.CreateFrame('Text', 'TwinABTBuffDetails.Caster', buffDetails)
	buffDetails.description:SetPoint('TOPLEFT', buffDetails.caster, 'BOTTOMLEFT', 0, 5)
	buffDetails.description:SetWidth(240)
	buffDetails.description:SetFontColor(1,1,1)
	buffDetails.description:SetFontSize(13)
	buffDetails.description:SetWordwrap(true)
	
	return buffDetails
end

-- Removes buff from sorted list if it is already added. Auras are ignored
local function removeBuffFromList(sortedList, buffID, aura)
	for i,id in ipairs(sortedList) do
		if id == buffID then
			if not aura then
				table.remove(sortedList, i)
				return true
			else
				return true
			end
		end
	end
	-- No buffs removed
	return false
end

-- Adds buff to proper spot in sorted list
local function addBuffToList(sortedList, list, buff)
	local isAura = not buff.duration
	local added = false
	
	-- Looking for possible duplicates
	local inList = removeBuffFromList(sortedList, buff.id, true)
	-- Looking for spot in the list
	if not isAura or isAura and not inList then
		for i,id in ipairs(sortedList) do
			if isAura then
				-- looking for first non-aura and adding to its spot
				if list[id].duration then 
					table.insert(sortedList, i, buff.id) 
					added = true
					break 
				end
			else
				removeBuffFromList(sortedList, buff.id)
				-- Comparing based on remaining time
				if list[id].duration and buff.remaining >= list[id].remaining then 
					table.insert(sortedList, i, buff.id) 
					added = true
					break 
				end
			end
		end
	else
		added = true
	end
	-- No place found, adding to the end of the list
	if not added then table.insert(sortedList, buff.id) end
end

function init()
	table.insert(Event.Addon.SavedVariables.Load.End, {function(identifier) if identifier == "TwinABT" then tABT:manageSavedVariables() end end, "TwinABT", "Saved variables loaded"})
end

function tABT:manageSavedVariables()
	if not tABTbuffmanager then tABTbuffmanager = {} end
	-- Buff row limit
	if not tABTbuffmanager.limitBuffInRow then tABTbuffmanager.limitBuffInRow = false end
	if not tABTbuffmanager.buffInRow then tABTbuffmanager.buffInRow = 12 end
	-- Buff Details
	tABTbuffmanager.buffDetails = tABT:GetBuffDetailsFrame()
	-- Drawing current buffs
	tABTbuffmanager.DrawBuffs = DrawBuffs
	-- Buff management
	tABTbuffmanager.addBuffToList = addBuffToList
	tABTbuffmanager.removeBuffFromList = removeBuffFromList
	tABTbuffmanager.MouseOver = MouseOver
end

init()