local tABT = {}
context = context or UI.CreateContext('TwinABTContext')
playerID = playerID or Inspect.Unit.Lookup('player')

tABTgroupsDefaults = {
	{
		name = "Archon",
		description = "Archon's Auras",
		icon = {"TwinABT", "archonAuras.png"},
		enabled = {
			["player"] = false,
			["group"] = false
		},
		buffs = {
			["Shared Vigor"] = true,
			["Tempered Armor"] = true,
			["Vitality of Stone"] = true,
			["Arcane Aegis"] = true,
			["Burning Purpose"] = true,
		},
		order = 1,
		["player"] = {},
		["group"] = {},
	},
	{
		name = "Motifs",
		description = "Bard's Motifs",
		icon = {"TwinABT", "motifs.png"},
		enabled = {
			["player"] = false,
			["group"] = false
		},
		buffs = {
			["Motif of Bravery"] = true,
			["Motif of Focus"] = true,
			["Motif of Grandeur"] = true,
			["Motif of Regeneration"] = true,
			["Motif of Renewal"] = true,
			["Motif of Tenacity"] = true,
			["Resonance"] = true
		},
		order = 2,
		["player"] = {},
		["group"] = {},
	},
	{
		name = "Auras",
		description = "Aura-type (without duration) buffs",
		icon = {"TwinABT", "auras.png"},
		enabled = {
			["player"] = false,
			["group"] = false
		},
		order = 3,
		["player"] = {},
		["group"] = {},
	},
	{
		name = "LongBuffs",
		description = "Buffs with remaining time >5min",
		icon = {"TwinABT", "longBuff.png"},
		enabled = {
			["player"] = false,
			["group"] = false
		},
		order = 4,
		["player"] = {},
		["group"] = {},
	},
}

-- Checks if mouse is inside given frame
local function MouseInFrame(frame)
	local mouse = Inspect.Mouse()
	local frameX1, frameX2 = frame:GetLeft(), frame:GetLeft()+frame:GetWidth()
	local frameY1, frameY2 = frame:GetTop(), frame:GetTop()+frame:GetHeight()
	
	return (frameX1-mouse.x <= 0) and (frameX2-mouse.x >= 0) and (frameY1-mouse.y <= 0) and (frameY2-mouse.y >= 0)
end

-- Returns a proper time format and suffix for buff remaining time
local function getTimeString(remaining)
	local rTime, rSuffix
	if remaining <= 60 then
		rTime = math.floor(remaining)
		rSuffix = 's'
	elseif remaining <= 3600 then
		rTime = math.floor(remaining / 60)
		rSuffix = 'm'
	else
		rTime = math.floor(remaining / 3600)
		rSuffix = 'h'
	end
	
	return rTime, rSuffix
end

-- Retrieves all viable buffs from the given list
local function RetrieveAll(group, playerList, groupList)
	if group.name == "Archon" or group.name == "Motifs" then
		-- Player buffs
		if group.enabled["player"] then
			-- finding all matching buffs
			for id,buff in pairs(playerList) do
				if group.buffs[buff.name] then
					group["player"][buff.id] = buff
				end
			end
			-- Removing all matched buffs
			for id,_ in pairs(group["player"]) do
				playerList[id] = nil
			end
		end
		-- Group buffs
		if group.enabled["group"] then
			-- finding all matching buffs
			for id,buff in pairs(groupList) do
				if group.buffs[id] then
					group["group"][id] = buff
				end
			end
			-- Removing all matched buffs
			for id,_ in pairs(group["group"]) do
				groupList[id] = nil
			end
		end
	elseif group.name == "Auras" then
		-- Player buffs
		if group.enabled["player"] then
			-- Finding all buffs matching Aura's definition
			for id, buff in pairs(playerList) do
				if not buff.duration then
					group["player"][id] = buff
				end
			end
			-- Removing all matched buffs
			for id,_ in pairs(group["player"]) do
				playerList[id] = nil
			end
		end
		-- Group buffs
		if group.enabled["group"] then
			-- finding all buffs matching Aura's definition
			for id,buff in pairs(groupList) do
				if not buff.duration then
					group["group"][id] = buff
				end
			end
			-- Removing all matched buffs
			for id,_ in pairs(group["group"]) do
				groupList[id] = nil
			end
		end
	elseif group.name == "LongBuffs" then
		-- Player buffs
		if group.enabled["player"] then
			-- Finding all Long buffs
			for id, buff in pairs(playerList) do
				if buff.remaining and buff.remaining > 300 then
					group["player"][id] = buff
				end
			end
			-- Removing all matched buffs
			for id,_ in pairs(group["player"]) do
				playerList[id] = nil
			end
		end
		-- Group buffs
		if group.enabled["group"] then
			-- finding all Long buffs
			for id,buff in pairs(groupList) do
				if buff.remaining and buff.remaining > 300 then
					group["group"][id] = buff
				end
			end
			-- Removing all matched buffs
			for id,_ in pairs(group["group"]) do
				groupList[id] = nil
			end
		end
	end
end

-- Checks if given buff matches any group
local function CheckBuff(group, buff, buffGroup)
	if group.name == "Archon" or group.name == "Motifs" then
		if group.enabled[buffGroup] then
			if group.buffs[buff.name] then
				group[buffGroup][buff.id] = buff
				return true
			end
		end
	elseif group.name == "Auras" then
		if group.enabled[buffGroup] then
			if not buff.duration then
				group[buffGroup][buff.id] = buff
				return true
			end
		end
	elseif group.name == "LongBuffs" then
		if group.enabled[buffGroup] then
			if buff.remaining and buff.remaining > 300 then
				group[buffGroup][buff.id] = buff
				return true
			end
		end
	end
end

-- Removes buffs from the list
local function RemoveBuff(buffID)
	local buffGroups = {"player", "group"}
	for i,group in ipairs(tABTgroups) do
		for _,buffGroup in ipairs(buffGroups) do
			if group.enabled[buffGroup] and group[buffGroup][buffID] then
				group[buffGroup][buffID] = nil
				return buffGroup
			end
		end
	end
	return false
end

-- Checks 'LongBuffs' group and returns buffs that doesn't match the criteria anymore
local function CheckLongBuffs(buffGroup)
	local outBuffs = {}
	for _,group in ipairs(tABTgroups) do
		if group.name == "LongBuffs" then
			for id,buff in pairs(group[buffGroup]) do
				local buffData = Inspect.Buff.Detail('player', id)
				if group[buffGroup][buffData.id] and buffData.remaining < 300 then
					table.insert(outBuffs,buffData)
					group[buffGroup][buffData.id] = nil
				else
					group[buffGroup][buffData.id] = buffData
				end
			end
		end
	end
	return outBuffs
end

-- Adds buff Group boxes to provided content
local function AddGroupsToContent(content, buffGroup)
	content.boxes = content.boxes or {}
	for i,group in ipairs(tABTgroups) do
		table.insert(content.boxes, tABT:CreateGroupBox(group, content, i, buffGroup))
	end
end

-- Updates Group boxes. Sets visibility and width of the content
local function UpdateGroupsContent(buffGroup, content)
	local count = 0
	-- Updating boxes
	for i,group in ipairs(tABTgroups) do
		if group.enabled[buffGroup] then
			-- Getting buffs in group
			local buffCount = 0
			for k,_ in pairs(group[buffGroup]) do
				buffCount = buffCount + 1
			end
			-- showing group only if there are buffs
			if buffCount > 0 then			
				-- Updating visibility
				content.boxes[i]:SetVisible(true)
				content.boxes[i].count:SetText(''..buffCount)
				-- Updating position
				local offsetX = (content.boxes[i]:GetWidth() + 1) * count  * (-1)
				content.boxes[i]:ClearPoint("TOPRIGHT")
				content.boxes[i]:SetPoint("TOPRIGHT", content, "TOPRIGHT", offsetX, 0)
				count = count + 1
				-- Redrawing group details
				if tABTgroups.groupDetails:GetVisible() and tABTgroups.groupDetails.group == group.name then
					tABTgroups.groupDetails.update()
				end
			else
				if tABTgroups.groupDetails.buffGroup == buffGroup and tABTgroups.groupDetails.group == group.name then 
					tABTgroups.groupDetails:SetVisible(false) 
				end
				content.boxes[i]:SetVisible(false)
			end
		else
			if tABTgroups.groupDetails.buffGroup == buffGroup and tABTgroups.groupDetails.group == group.name then 
				tABTgroups.groupDetails:SetVisible(false) 
			end
			content.boxes[i]:SetVisible(false)
		end
	end
	-- Updating Content
	local boxWidthWithSpacing = content.boxes[1]:GetWidth() + 1
	local newWidth = boxWidthWithSpacing * count
	content:SetWidth(newWidth)
end

-- Resets all Buff lists of the groups
local function ResetGroups()
	for _,group in ipairs(tABTgroups) do
		group["player"] = {}
		group["group"] = {}
	end
end

-- Creates and returns a Group Box
function tABT:CreateGroupBox(group, parent, position, buffGroup)
	local boxConfig = {
		width = 30,
		height = 30,
		border = 1,
		count = {
			fontSize = 13,
			fontColor = {1,1,1,1},
			layer = 2,
			height = 20,
			background = {0, 0, 0, 0.8}
		},
	}

	-- Border
	local groupBox = UI.CreateFrame('Frame', 'TwinABT Group Box', parent)
	groupBox:SetPoint('TOPRIGHT', parent, 'TOPRIGHT', 0, 0)
	groupBox:SetWidth(boxConfig.width)
	groupBox:SetHeight(boxConfig.height)
	groupBox:SetBackgroundColor(0,0,0)
	groupBox:SetVisible(false)
	-- Texture
	local groupTexture = UI.CreateFrame('Texture', 'TwinABT Group Box Texture', groupBox)
	groupTexture:SetPoint('TOPLEFT', groupBox, 'TOPLEFT', 1, 1)
	groupTexture:SetWidth(boxConfig.width - (boxConfig.border * 2))
	groupTexture:SetHeight(boxConfig.height - (boxConfig.border * 2))
	groupTexture:SetTexture(unpack(group.icon))
	-- Count
	local groupCount = UI.CreateFrame('Text', 'TwinABT Group Box Count', groupBox)
	groupCount:SetPoint('CENTERTOP', groupBox, 'CENTERBOTTOM', 0, 1)
	groupCount:SetLayer(boxConfig.count.layer)
	groupCount:SetFontSize(boxConfig.count.fontSize)
	groupCount:SetFontColor(unpack(boxConfig.count.fontColor))
	groupCount:SetText('2')
	-- Count Background
	local groupCountBG = UI.CreateFrame('Frame', 'TwinABT Group Box Count Background', groupBox)
	groupCountBG:SetPoint('CENTERTOP', groupBox, 'CENTERBOTTOM', 0, 0)
	groupCountBG:SetLayer(boxConfig.count.layer-1)
	groupCountBG:SetHeight(boxConfig.count.height)
	groupCountBG:SetWidth(boxConfig.width)
	groupCountBG:SetBackgroundColor(unpack(boxConfig.count.background))
	-- Keeping track
	groupBox.texture = groupTexture
	groupBox.count = groupCount
	groupBox.countBG = groupCountBG
	groupBox.group = group
	groupBox.buffGroup = buffGroup
	
		-- Management
	groupBox.ShowDetails = function(self, click, forcedFocus)
		tABTgroups.groupDetails:ClearPoint("TOPRIGHT")
		tABTgroups.groupDetails:SetPoint("TOPRIGHT", self, "BOTTOMRIGHT", 0, 0)
		tABTgroups.groupDetails:SetVisible(true)
		tABTgroups.groupDetails.group = self.group.name
		tABTgroups.groupDetails.buffGroup = self.buffGroup
		tABTgroups.groupDetails.groupBox = self
		tABTgroups.groupDetails.showBoxes(self.group[self.buffGroup])
		if click then
			tABTgroups.groupDetails.focus = forcedFocus or not tABTgroups.groupDetails.focus
			tABTgroups.groupDetails.time = os.time()
		end
	end
	
	-- Mouse Over: Buff details
	function groupBox.Event:MouseIn()
		self.ShowDetails(self)
		--if self.buffID then tABT:MouseOver(self) end
	end
	
	-- Mouse Out: hide buff details
	function groupBox.Event:MouseOut()
		if tABTgroups.groupDetails.group == self.group.name and not tABTgroups.groupDetails.focus then 
			tABTgroups.groupDetails:SetVisible(false) 
			tABTgroups.groupDetails.focus = false
		end
	end
	
	-- Mouse Right Click: remove buff
	function groupBox.Event:LeftClick()
		self.ShowDetails(self, true)
	end
	
	-- Adding new box to list
	return groupBox
end

-- Creates and returns a Buff Box
local function CreateBuffBox(parent)
	local boxConfig = {
		width = 30,
		height = 30,
		spacing = 1,
		border = 1,
		text = {
			fontSize = 13,
			fontColor = {1,1,1,1},
			layer = 2,
			height = 20,
			background = {0, 0, 0, 0.8}
		},
		stacks = {
			fontSize = 12,
			fontColor = {1,1,1,1},
			layer = 2,
			background = {0,0,0,0.9}
		},
	}
	-- Border
	local buffBox = UI.CreateFrame('Frame', 'TwinABT Buffs Box', parent)
	buffBox:SetPoint('TOPRIGHT', parent, 'TOPRIGHT', 0, 0)
	buffBox:SetWidth(boxConfig.width)
	buffBox:SetHeight(boxConfig.height)
	buffBox:SetBackgroundColor(0,0,0)
	buffBox:SetVisible(false)
	-- Texture
	local buffTexture = UI.CreateFrame('Texture', 'TwinABT Buffs Box', buffBox)
	buffTexture:SetPoint('TOPLEFT', buffBox, 'TOPLEFT', 1, 1)
	buffTexture:SetWidth(boxConfig.width - (boxConfig.border * 2))
	buffTexture:SetHeight(boxConfig.height - (boxConfig.border * 2))
	-- Text
	local boxText = UI.CreateFrame('Text', 'TwinABT Buffs Box Text', buffBox)
	boxText:SetPoint('CENTERTOP', buffBox, 'CENTERBOTTOM', 0, 1)
	boxText:SetLayer(boxConfig.text.layer)
	boxText:SetFontSize(boxConfig.text.fontSize)
	boxText:SetFontColor(unpack(boxConfig.text.fontColor))
	-- Text Background
	local boxTextBG = UI.CreateFrame('Frame', 'TwinABT Buffs Box Text Background', buffBox)
	boxTextBG:SetPoint('CENTERTOP', buffBox, 'CENTERBOTTOM', 0, 0)
	boxTextBG:SetLayer(boxConfig.text.layer-1)
	boxTextBG:SetHeight(boxConfig.text.height)
	boxTextBG:SetWidth(boxConfig.width)
	boxTextBG:SetBackgroundColor(unpack(boxConfig.text.background))
	-- Stacks
	local boxStacks = UI.CreateFrame('Text', 'TwinABT Buffs Box Stacks', buffBox)
	boxStacks:SetPoint('TOPRIGHT', buffBox, 'TOPRIGHT', 0, 0)
	boxStacks:SetLayer(boxConfig.stacks.layer)
	boxStacks:SetFontSize(boxConfig.stacks.fontSize)
	boxStacks:SetFontColor(unpack(boxConfig.stacks.fontColor))
	boxStacks:SetAlpha(0)
	-- Stacks shadow
	local boxStacksBG = UI.CreateFrame('Frame', 'TwinABT Buffs Box Stacks Background', buffBox)
	boxStacksBG:SetPoint('CENTER', boxStacks, 'CENTER', 0, 0)
	boxStacksBG:SetLayer(boxConfig.stacks.layer - 1)
	boxStacksBG:SetBackgroundColor(unpack(boxConfig.stacks.background))
	boxStacksBG:SetAlpha(0)
	-- Keeping track
	buffBox.texture = buffTexture
	buffBox.text = boxText
	buffBox.textBG = boxTextBG
	buffBox.stacks = boxStacks
	buffBox.stacksBG = boxStacksBG
	
	-- Mouse Over: Buff details
	function buffBox.Event:MouseIn()
		if self.buffID and tABTgroups.groupDetails.focus then MouseOver(self) end
	end
	
	-- Mouse Out: hide buff details
	function buffBox.Event:MouseOut()
		if tABTbuffmanager.buffDetails.buff == self.buffID then tABTbuffmanager.buffDetails:SetVisible(false) end
	end
	
	-- Mouse Right Click: remove buff
	function buffBox.Event:RightClick()
		if self.buffID then
			-- Hiding buff details
			if buffDetails.buff == self.buffID then buffDetails:SetVisible(false) end
			-- Canceling buff
			Command.Buff.Cancel(self.buffID) 
		end
	end
	
	-- Adding new box to list
	return buffBox
end

-- Resets the given box
function tABT:ResetBox(box)
	box:SetVisible(false)
	box.buffID = nil
	-- time
	box.text:SetAlpha(0)
	box.textBG:SetAlpha(0)
	-- stacks
	box.stacks:SetAlpha(0)
	box.stacksBG:SetAlpha(0)
end

-- Creates and returns a group details frame
function tABT:GetGroupDetailsFrame()
	-- Group Details
	groupDetails = UI.CreateFrame('Frame', 'GroupDetails', context)
	groupDetails:SetWidth(250)
	groupDetails:SetHeight(60)
	groupDetails:SetLayer(1)
	groupDetails:SetBackgroundColor(0,0,0,0.9)
	groupDetails:SetVisible(false)
	-- Inside boxes
	groupDetails.boxes = {}
	groupDetails.showBoxes = function(buffs)
			-- Sorting buffs list if it has remaining time
			local sortedBuffs = {}
			for _,buff in pairs(buffs) do table.insert(sortedBuffs, Inspect.Buff.Detail("player", buff.id)) end
			if sortedBuffs[1].remaining then
				table.sort(sortedBuffs, function(a, b) 
					if not a or not b then return false 
					else return a.remaining > b.remaining end
				end)
			end
			buffs = sortedBuffs
			-- showing boxes
			local hasTime = false
			local count = 0 -- total count of buffs
			local layer = 1 -- number of rows
			local position = 0 -- buff position in row
			for _,buff in ipairs(sortedBuffs) do
				-- refreshing buff data
				buff = Inspect.Buff.Detail('player', buff.id)
				count = count + 1
				-- Checking if there are enough boxes
				if not groupDetails.boxes[count] then groupDetails.boxes[count] = CreateBuffBox(groupDetails) end
				-- Reseting box
				tABT:ResetBox(groupDetails.boxes[count])
				-- updating box to match buff
				
				local offsetX = (position * (-31))-5
				local offsetY = 5 + ((hasTime and 51 or 31) * (layer-1))
				position = position + 1
				if tABTbuffmanager.limitBuffInRow and count~=1 and count%tABTbuffmanager.buffInRow == 0 then
					position = 0
					layer = layer+1
				end
				--[[
				--* Breakpoint (buffs in row limit reached)
				if tABTbuffmanager.limitBuffInRow and i~=1 and (i-1)%tABTbuffmanager.buffInRow == 0 then
					groupDetails.boxes[count]:SetPoint("TOPRIGHT", groupDetails, "TOPRIGHT", offsetX, 5)
					
					content.boxes[i]:SetPoint("TOPRIGHT", content.boxes[1+(tABTbuffmanager.buffInRow*(((i-1)/tABTbuffmanager.buffInRow)-1))], "BOTTOMRIGHT", 0, 21)
					content.boxes[i].breakpoint = true
				end]]--
				
				--local offsetX = ((count-1) * (-31))-5
				groupDetails.boxes[count]:ClearPoint("TOPRIGHT")
				groupDetails.boxes[count]:SetPoint("TOPRIGHT", groupDetails, "TOPRIGHT", offsetX, offsetY)
				groupDetails.boxes[count].texture:SetTexture('Rift', buff.icon)
				groupDetails.boxes[count]:SetVisible(true)
				groupDetails.boxes[count].buffID = buff.id
				-- Remaining time
				if buff.remaining then
					local rTime, rSuffix = getTimeString(buff.remaining)
					-- Setting text
					groupDetails.boxes[count].text:SetAlpha(1)
					groupDetails.boxes[count].text:SetText(rTime .. rSuffix)
					groupDetails.boxes[count].textBG:SetAlpha(1)
					hasTime = true
				end
				-- Stacks
				if buff.stack then
					groupDetails.boxes[count].stacks:SetText('' .. buff.stack)
					groupDetails.boxes[count].stacks:SetAlpha(1)
					groupDetails.boxes[count].stacksBG:SetAlpha(1)
					groupDetails.boxes[count].stacksBG:SetHeight(groupDetails.boxes[count].stacks:GetHeight())
					groupDetails.boxes[count].stacksBG:SetWidth(groupDetails.boxes[count].stacks:GetWidth())
				end
			end
			groupDetails.hasTime = hasTime
			-- updating length and height
			if not tABTbuffmanager.limitBuffInRow then
				local boxesWidth = (count * 31)+9
				groupDetails:SetWidth(boxesWidth)
				groupDetails:SetHeight(hasTime and 60 or 40)
			else
				-- width
				if count < tABTbuffmanager.buffInRow then
					local boxesWidth = (count * 31)+9
					groupDetails:SetWidth(boxesWidth)
				else
					local boxesWidth = (tABTbuffmanager.buffInRow * 31)+9
					groupDetails:SetWidth(boxesWidth)
				end
				--height
				if count%tABTbuffmanager.buffInRow == 0 then layer = layer - 1 end
				local boxesHeight = ((hasTime and 50 or 30) * layer) + (1*(layer-1)) + 10
				groupDetails:SetHeight(boxesHeight)
			end
			-- reseting unused boxes
			if count+1 <= #groupDetails.boxes then
				for i=count+1, #groupDetails.boxes do
					tABT:ResetBox(groupDetails.boxes[i])
				end
			end
		end
	-- updating inside boxes
	groupDetails.update = function()
		-- redrawing list
		groupDetails.groupBox.ShowDetails(groupDetails.groupBox, tABTgroups.groupDetails.focus, true)
	end
	
	-- Updates remaining time of displayed buffs
	groupDetails.updateTime = function()
		for i,box in ipairs(groupDetails.boxes) do
			if box.buffID then
				local buff = Inspect.Buff.Detail("player", box.buffID)
				local rTime, rSuffix = getTimeString(buff.remaining)
				box.text:SetText(rTime .. rSuffix)
			end
		end
	end
	
	-- Box functionality
	function groupDetails.Event:MouseOut()
		if self.focus then
			-- Checking if mouse has left the box
			if not MouseInFrame(self) then
				self:SetVisible(false)
				self.focus = false
			end
		end
	end
	
	return groupDetails
end

function init()
	table.insert(Event.Addon.SavedVariables.Load.End, {function(identifier) if identifier == "TwinABT" then tABT:manageSavedVariables() end end, "TwinABT", "Saved variables loaded"})
end

function tABT:manageSavedVariables()
	if not tABTgroups then tABTgroups = tABTgroupsDefaults end
	tABTgroups.groupDetails = tABT:GetGroupDetailsFrame()
	-- adding functionality
	tABTgroups.RetrieveAll = RetrieveAll
	tABTgroups.AddGroupsToContent = AddGroupsToContent
	tABTgroups.UpdateGroupsContent = UpdateGroupsContent
	tABTgroups.ResetGroups = ResetGroups
	tABTgroups.CheckBuff = CheckBuff
	tABTgroups.RemoveBuff = RemoveBuff
	tABTgroups.CheckLongBuffs = CheckLongBuffs
	
end

init()