--[[
	Options panel for TwinABT
]]--
local optionsContext = UI.CreateContext('TwinABTOptionsContext')

local tABT = {}

tABToptionsDefaults = {
	toggleButton = {
		texture = "icon.png",
		textureHover = "iconHover.png",
		textureFocus = "iconPress.png",
		width = 30,
		height = 30,
	},
	mainFrame = {
		width = 400,
		height = 500,
		visible = false,
		title = "Twin: Advanced Buff Tracker",
	},
	anchors = {
		buffs = {},
		debuffs = {}
	}
}

function init()
	--
	table.insert(Event.Addon.SavedVariables.Load.End, {function(identifier) if identifier == "TwinABT" then tABT:manageSavedVariables() end end, "TwinABT", "Saved variables loaded"})
end

-- Creates a text shadow
local function dropShadow(textFrame)
	local shadow = UI.CreateFrame("Text", "Text Shadow", textFrame:GetParent())
	shadow:SetPoint("TOPLEFT", textFrame, "TOPLEFT", 1, 1)
	shadow:SetFontSize(textFrame:GetFontSize())
	shadow:SetWidth(textFrame:GetWidth())
	shadow:SetWordwrap(true)
	shadow:SetText(textFrame:GetText())
	shadow:SetFontColor(0,0,0)
	shadow:SetLayer(textFrame:GetLayer() - 1)
	
	textFrame.shadow = shadow
end

-- Creating Options Panel
function tABT:CreateOptionsFrame()
	-- Creating Window
	frame = UI.CreateFrame("RiftWindow", "TwinABT Options Main Frame", optionsContext)
	frame:SetWidth(tABToptions.mainFrame.width)
	frame:SetHeight(tABToptions.mainFrame.height)
	frame:SetVisible(tABToptions.mainFrame.visible)
	frame:SetTitle(tABToptions.mainFrame.title)
	if not tABToptions.mainFrame.x then
		tABToptions.mainFrame.x = (UIParent:GetWidth() - tABToptions.mainFrame.width) / 2
		tABToptions.mainFrame.y = (UIParent:GetHeight() - tABToptions.mainFrame.height) / 2
	end
	frame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", tABToptions.mainFrame.x, tABToptions.mainFrame.y)
	-- Keeping track
	tABT.mainFrame = frame
	tABT.mainFrame.content = frame:GetContent()
	-- global
	tABToptions.mainFrame.frame = frame
	
	
	-- Adding Window elements:
	-- Close Button
	closeBtn = UI.CreateFrame("RiftButton", "Twin ABT Options Main Frame Close Options", tABT.mainFrame)
	closeBtn:SetSkin("close")
	closeBtn:SetPoint("TOPRIGHT", tABT.mainFrame, "TOPRIGHT", -8, 18)
	closeBtn:SetText("Close")
	closeBtn:SetLayer(1)
	function closeBtn.Event:LeftClick()
		tABT.mainFrame:SetVisible(false)
		tABTbuffs.EnableEditMode(false)
	end
	-- Header (drag frame)
	local border = tABT.mainFrame:GetBorder()
	local content = tABT.mainFrame.content
	header = UI.CreateFrame("Frame", "Twin ABT Options Main Frame Header", tABT.mainFrame)
	header:SetPoint("TOPLEFT", tABT.mainFrame, "TOPLEFT")
	header:SetWidth(border:GetWidth())
	header:SetHeight(content:GetTop() - border:GetTop())
	tABT.mainFrame.header = header
	-- Dragging
	function header.Event:LeftDown()
		self.drag = true
	end
	function header.Event:LeftUp()
		self.drag = false
		self.offset = nil
		tABToptions.mainFrame.x = self:GetLeft()
		tABToptions.mainFrame.y = self:GetTop()
	end
	function header.Event:MouseMove(x,y)
		if self.drag then
			if not self.offset then
				self.offset = {
					x = x-self:GetLeft(),
					y = y-self:GetTop()
				}
			end
			local moveX, moveY = x-self.offset.x, y-self.offset.y
			tABT.mainFrame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", moveX, moveY)
		end
	end
	
end

-- Creating tabs for options
function tABT:CreateTabController()
	local Content = tABT.mainFrame.content
	
	-- Controller
	if not tABT.tabController then tABT.tabController = {} end

	-- Tab bars holder
	local tabsHolder = UI.CreateFrame("Frame", "TwinABT Options Tabs Holder", Content)
	tabsHolder:SetPoint("CENTERTOP", Content, "CENTERTOP", 0, 5)
	tabsHolder:SetWidth(Content:GetWidth()-10)
	tabsHolder:SetHeight(38)
	tABT.tabController.tabsHolder = tabsHolder
	
	-- Content
	local tabsContentHolder = UI.CreateFrame("Mask", "TwinABT Options Tabs Content Holder", Content)
	tabsContentHolder:SetPoint("CENTERTOP", tabsHolder, "CENTERBOTTOM", 0, 0)
	tabsContentHolder:SetWidth(Content:GetWidth()-10)
	tabsContentHolder:SetHeight(Content:GetHeight() - tabsHolder:GetHeight() - 10)
	tabsContentHolder:SetBackgroundColor(1,1,1,0.2)
	tABT.tabController.tabsContentHolder = tabsContentHolder
	
	-- Controller functions
	--* Add tab item
	tABT.tabController.AddTabItem = function(title)
		if not tABT.tabController.items then tABT.tabController.items = {} end
		-- Tabs holder
		local tabHolder = tABT.tabController.tabsHolder
		-- Tab item
		--* Background
		local item = UI.CreateFrame("Frame", "TwinABT Options Tab Item", tabHolder)
		item:SetHeight(tabHolder:GetHeight())
		item:SetBackgroundColor(0,0,0,0.5)
		if #tABT.tabController.items == 0 then
			item:SetPoint("CENTERLEFT", tabHolder, "CENTERLEFT", 1, 0)
		else
			item:SetPoint("CENTERLEFT", tABT.tabController.items[#tABT.tabController.items], "CENTERRIGHT", 1, 0)
		end
		--* Title
		item.title = UI.CreateFrame("Text", "TwinABT Options Tab Item Title", frame)
		item.title:SetText(title)
		item.title:SetFontSize(14)
		item.title:SetFontColor(237/255,191/255,47/255)
		item.title:SetPoint("CENTER", item, "CENTER", 0, 0)
		dropShadow(item.title)
		
		item:SetWidth(item.title:GetWidth() + 10)
		--* Events
		function item.Event:MouseIn()
			if not self.selected then
				self:SetBackgroundColor(1,1,1,0.2)
			end
		end
		
		function item.Event:MouseOut()
			if not self.selected then
				self:SetBackgroundColor(0,0,0,0.5)
			end
		end
		
		function item.Event:LeftClick()
			if not self.selected then
				tABT.tabController.selectItem(self)
			end
		end
		
		item.setSelected = function(self)
			self:SetBackgroundColor(1,1,1,0.2)
			self.title:SetFontColor(1,1,1,1)
			self.selected = true
		end
		
		item.clearSelected = function(self)
			self:SetBackgroundColor(0,0,0,0.5)
			self.title:SetFontColor(237/255,191/255,47/255)
			self.selected = false
		end
		
		--* Keeping track
		table.insert(tABT.tabController.items, item)
		
		--* Returning
		return item
	end
	
	--* Select tab item
	tABT.tabController.selectItem = function(item)
		-- if tab position and not actual item was passed
		if tABT.tabController.items[item] then item = tABT.tabController.items[item] end
		
		-- unselecting previous item
		if tABT.tabController.selectedItem then
			tABT.tabController.selectedItem:clearSelected()
			tABT.tabController.selectedItem.content:SetVisible(false)
		end
		
		-- selecting the new item
		tABT.tabController.selectedItem = item
		item:setSelected()
		item.content:SetVisible(true)
		
	end
	
	
end

-- Creates Tabs and fills their content
function tABT:AddTabItems()
	local Content = tABT.tabController.tabsContentHolder
	
	-- General options
	local tab1 = tABT.tabController.AddTabItem("General")
	--* Adding content
	tab1.content = UI.CreateFrame("Frame", "TwinABT Options Tab1 Content", Content)
	tab1.content:SetPoint("CENTERTOP", Content, "CENTERTOP", 0, 5)
	tab1.content:SetWidth(Content:GetWidth()-10)
	tab1.content:SetHeight(Content:GetHeight()-10)
	tab1.content:SetVisible(false)
	--* Filling content
	tABT:FillGeneralContent(tab1.content)
	
	-- Buff Groups options
	local tab2 = tABT.tabController.AddTabItem("Buff Groups")
	--* Adding content
	tab2.content = UI.CreateFrame("Frame", "TwinABT Options Tab2 Content", Content)
	tab2.content:SetPoint("CENTERTOP", Content, "CENTERTOP", 0, 5)
	tab2.content:SetWidth(Content:GetWidth()-10)
	tab2.content:SetHeight(Content:GetHeight()-10)
	tab2.content:SetVisible(false)
	--* Filling content
	tABT:FillBuffGroupsContent(tab2.content)
	
	tABT.tabController.selectItem(1)
end

-- Filling General Options content
function tABT:FillGeneralContent(Content)
	Content.items = {}
	-- Buffs in a row
	local frame1 = UI.CreateFrame("Frame", "TwinABT Options General Frame1", Content)
	frame1:SetPoint("TOPLEFT", Content, "TOPLEFT", 0, 0)
	frame1:SetWidth(Content:GetWidth())
	frame1:SetHeight(140)
	frame1:SetBackgroundColor(0,0,0,0.6)
	frame1.fillElements = function(frame)
		--* Checkbox
		frame.chk1 = UI.CreateFrame("RiftCheckbox", "TwinABT Options General CheckBox1", frame)
		frame.chk1:SetPoint("TOPLEFT", frame, "TOPLEFT", 5, 5)
		frame.chk1:SetChecked(tABTbuffmanager.limitBuffInRow)
		frame.chk1:SetEnabled(true)
		frame.chk1.parent = frame
		function frame.chk1.Event:CheckboxChange()
			self.parent.slider:SetEnabled(self:GetChecked())
			tABTbuffmanager.limitBuffInRow = self:GetChecked()
		end
		---* Checkbox Text
		frame.chk1.text = UI.CreateFrame("Text", "TwinABT Options General CheckBox1 Text", frame)
		frame.chk1.text:SetPoint("CENTERLEFT", frame.chk1, "CENTERRIGHT", 2, 0)
		frame.chk1.text:SetText("Limit buffs in a row")
		frame.chk1.text:SetFontSize(16)
		frame.chk1.text:SetFontColor(237/255,191/255,47/255)
		dropShadow(frame.chk1.text)
		
		--* Slider
		frame.slider = UI.CreateFrame("RiftSlider", "TwinABT Options General Slider1", frame)
		frame.slider:SetPoint("CENTERTOP", frame, "CENTERTOP", 0, 55)
		frame.slider:SetWidth(Content:GetWidth()-20)
		frame.slider:SetRange(2,20)
		frame.slider:SetPosition(tABTbuffmanager.buffInRow)
		frame.slider:SetEnabled(tABTbuffmanager.limitBuffInRow)
		function frame.slider.Event:SliderChange()
			self.position:SetText("" .. self:GetPosition())
			self.position.shadow:SetText(self.position:GetText())
			tABTbuffmanager.buffInRow = self:GetPosition()
		end
		---* Slider: position
		frame.slider.position = UI.CreateFrame("Text", "TwinABT Options General Slider Position", frame)
		frame.slider.position:SetPoint("CENTERBOTTOM", frame.slider, "CENTERTOP", 0, -5)
		frame.slider.position:SetText("" .. frame.slider:GetPosition())
		frame.slider.position:SetFontSize(14)
		frame.slider.position:SetFontColor(1,1,1)
		dropShadow(frame.slider.position)
		frame.slider.position.shadow:SetWordwrap(false)
		
		---* Slider: Min/Max values
		local minimum, maximum = frame.slider:GetRange()
		
		frame.slider.min = UI.CreateFrame("Text", "TwinABT Options General Slider Position Min", frame)
		frame.slider.min:SetPoint("TOPLEFT", frame.slider, "BOTTOMLEFT", 0, -3)
		frame.slider.min:SetText("" .. minimum)
		frame.slider.min:SetFontSize(14)
		frame.slider.min:SetFontColor(237/255,191/255,47/255)
		dropShadow(frame.slider.min)
		
		frame.slider.max = UI.CreateFrame("Text", "TwinABT Options General Slider Position Max", frame)
		frame.slider.max:SetPoint("TOPRIGHT", frame.slider, "BOTTOMRIGHT", 0, -3)
		frame.slider.max:SetText("" .. maximum)
		frame.slider.max:SetFontSize(14)
		frame.slider.max:SetFontColor(237/255,191/255,47/255)
		dropShadow(frame.slider.max)
		
		--* Note
		frame.note = UI.CreateFrame("Text", "TwinABT Options General CheckBox1 Note", frame)
		frame.note:SetText("Note: The limit does not accommodate the number of visible groups.")
		frame.note:SetFontSize(13)
		frame.note:SetFontColor(1,1,1)
		frame.note:SetWidth(frame:GetWidth()-10)
		frame.note:SetWordwrap(true)
		frame.note:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT", 5, -5)
		dropShadow(frame.note)
	end
	--* Filling elements and keeping track
	frame1:fillElements()
	table.insert(Content.items, frame1)
	
	-- Extra info
	local frame2 = UI.CreateFrame("Frame", "TwinABT Options General Frame2", Content)
	frame2:SetPoint("TOPLEFT", frame1, "BOTTOMLEFT", 0, 5)
	frame2:SetWidth(Content:GetWidth())
	frame2:SetHeight(110)
	frame2:SetBackgroundColor(0,0,0,0.6)
	frame2.fillElements = function(frame)
		-- Title
		frame.title = UI.CreateFrame("Text", "TwinABT Options General Frame2 Title", frame)
		frame.title:SetPoint("TOPLEFT", frame, "TOPLEFT", 5, 5)
		frame.title:SetText("Extra Info - Debuff Coloring")
		frame.title:SetFontSize(16)
		frame.title:SetFontColor(237/255,191/255,47/255)
		dropShadow(frame.title)
		-- debuffs
		local colors = {
			disease = {250/255,229/255,36/255},
			poison = {59/255,172/255,2/255},
			curse = {159/255,4/255,242/255},
			other = {0,0,0}
		}
		--* disease
		frame.disease = UI.CreateFrame("Frame", "TwinABT Options General Frame2 Disease", frame)
		frame.disease:SetPoint("TOPLEFT", frame.title, "BOTTOMLEFT", 0, 3)
		frame.disease:SetWidth(15)
		frame.disease:SetHeight(15)
		frame.disease:SetBackgroundColor(unpack(colors.disease))
		frame.disease.text = UI.CreateFrame("Text", "TwinABT Options General Frame2 Disease Text", frame)
		frame.disease.text:SetPoint("CENTERLEFT", frame.disease, "CENTERRIGHT", 3, 0)
		frame.disease.text:SetText("Disease")
		frame.disease.text:SetFontSize(14)
		frame.disease.text:SetFontColor(unpack(colors.disease))
		dropShadow(frame.disease.text)
		--* poison
		frame.poison = UI.CreateFrame("Frame", "TwinABT Options General Frame2 Poison", frame)
		frame.poison:SetPoint("TOPLEFT", frame.disease, "BOTTOMLEFT", 0, 3)
		frame.poison:SetWidth(15)
		frame.poison:SetHeight(15)
		frame.poison:SetBackgroundColor(unpack(colors.poison))
		frame.poison.text = UI.CreateFrame("Text", "TwinABT Options General Frame2 Poison Text", frame)
		frame.poison.text:SetPoint("CENTERLEFT", frame.poison, "CENTERRIGHT", 3, 0)
		frame.poison.text:SetText("Poison")
		frame.poison.text:SetFontSize(14)
		frame.poison.text:SetFontColor(unpack(colors.poison))
		dropShadow(frame.poison.text)
		--* curse
		frame.curse = UI.CreateFrame("Frame", "TwinABT Options General Frame2 Curse", frame)
		frame.curse:SetPoint("TOPLEFT", frame.poison, "BOTTOMLEFT", 0, 3)
		frame.curse:SetWidth(15)
		frame.curse:SetHeight(15)
		frame.curse:SetBackgroundColor(unpack(colors.curse))
		frame.curse.text = UI.CreateFrame("Text", "TwinABT Options General Frame2 Curse Text", frame)
		frame.curse.text:SetPoint("CENTERLEFT", frame.curse, "CENTERRIGHT", 3, 0)
		frame.curse.text:SetText("Curse")
		frame.curse.text:SetFontSize(14)
		frame.curse.text:SetFontColor(unpack(colors.curse))
		dropShadow(frame.curse.text)
		--* other
		frame.other = UI.CreateFrame("Frame", "TwinABT Options General Frame2 Other", frame)
		frame.other:SetPoint("TOPLEFT", frame.curse, "BOTTOMLEFT", 0, 3)
		frame.other:SetWidth(15)
		frame.other:SetHeight(15)
		frame.other:SetBackgroundColor(unpack(colors.other))
		frame.other.text = UI.CreateFrame("Text", "TwinABT Options General Frame2 Other Text", frame)
		frame.other.text:SetPoint("CENTERLEFT", frame.other, "CENTERRIGHT", 3, 0)
		frame.other.text:SetText("Other")
		frame.other.text:SetFontSize(14)
		frame.other.text:SetFontColor(unpack(colors.other))
		dropShadow(frame.other.text)
		
	end
	--* Filling elements and keeping track
	frame2:fillElements()
	table.insert(Content.items, frame2)
end

-- Filing Buff Groups Options content
function tABT:FillBuffGroupsContent(Content)
	--local Content = tABT.mainFrame.content
	local contentHeight = 0
	local spacing = 5
	
	-- Groups
	Content.items = {}
	local lastTarget = Content
	for i,group in ipairs(tABTgroups) do
		local groupFrame = tABT:GetGroupSettingsFrame(group, Content, lastTarget)
		table.insert(Content.items, groupFrame)
		lastTarget = groupFrame
		contentHeight = contentHeight + groupFrame:GetHeight() + spacing
	end
	
	-- Fixing Content Height
	Content:SetHeight(contentHeight)
	--[[
	contentHeight = contentHeight + header:GetHeight() + (spacing*2)
	local heightFix = Content:GetHeight() - contentHeight
	tABT.mainFrame:SetHeight(tABT.mainFrame:GetHeight() - heightFix)
	]]--
end

-- Creates and returns options panel for group
function tABT:GetGroupSettingsFrame(group, parent, target)
	local frame = UI.CreateFrame("Frame", "TwinABT Options Group Settings", parent)
	if parent == target then
		frame:SetPoint("CENTERTOP", target, "CENTERTOP", 0, 0)
	else
		frame:SetPoint("CENTERTOP", target, "CENTERBOTTOM", 0, 5)
	end
	frame:SetWidth(parent:GetWidth())
	frame:SetHeight(40)
	frame:SetBackgroundColor(0,0,0,0.6)
	
	-- Icon
	local iconBorder = UI.CreateFrame("Frame", "TwinABT Options Group Settings Icon Border", frame)
	iconBorder:SetPoint("TOPLEFT", frame, "TOPLEFT", 5, 5)
	iconBorder:SetWidth(30)
	iconBorder:SetHeight(30)
	iconBorder:SetBackgroundColor(0,0,0)
	
	local icon = UI.CreateFrame("Texture", "TwinABT Options Group Settings Icon", iconBorder)
	icon:SetPoint("TOPLEFT", iconBorder, "TOPLEFT", 1, 1)
	icon:SetTexture(unpack(group.icon))
	icon:SetWidth(28)
	icon:SetHeight(28)
	icon.border = iconBorder
	
	-- Description
	local description = UI.CreateFrame("Text", "TwinABT Options Group Settings Description", frame)
	description:SetPoint("TOPLEFT", iconBorder, "TOPRIGHT", 5, 0)
	description:SetWidth(frame:GetWidth() - 45)
	description:SetWordwrap(true)
	description:SetText(group.description)
	description:SetFontSize(14)
	description:SetFontColor(237/255,191/255,47/255)
	dropShadow(description)
	
	-- Checkboxes
	local chk1 = UI.CreateFrame("RiftCheckbox", "TwinABT Options Group Settings CheckBox Self", frame)
	chk1:SetPoint("TOPLEFT", description, "BOTTOMLEFT", 0, 5)
	chk1:SetChecked(group.enabled["player"])
	chk1:SetEnabled(true)
	chk1.text = UI.CreateFrame("Text", "TwinABT Options Group Settings CheckBox Self Text", frame)
	chk1.text:SetPoint("CENTERLEFT", chk1, "CENTERRIGHT", 2, 0)
	chk1.text:SetText("Self")
	chk1.text:SetFontSize(14)
	chk1.text:SetFontColor(1,1,1)
	
	local chk2 = UI.CreateFrame("RiftCheckbox", "TwinABT Options Group Settings CheckBox Group", frame)
	chk2:SetPoint("TOPLEFT", chk1, "BOTTOMLEFT", 0, 5)
	chk2:SetChecked(group.enabled["group"])
	chk2:SetEnabled(true)
	chk2.text = UI.CreateFrame("Text", "TwinABT Options Group Settings CheckBox Group Text", frame)
	chk2.text:SetPoint("CENTERLEFT", chk2, "CENTERRIGHT", 2, 0)
	chk2.text:SetText("Group")
	chk2.text:SetFontSize(14)
	chk2.text:SetFontColor(1,1,1)
	
	-- Checboxes Event
	function chk1.Event:CheckboxChange()
		group.enabled["player"] = self:GetChecked()
	end
	function chk2.Event:CheckboxChange()
		group.enabled["group"] = self:GetChecked()
	end
	
	-- Updating Height
	frame:SetHeight(description:GetHeight() + chk1:GetHeight() + chk2:GetHeight() + 20)
	
	
	-- keeping track
	frame.icon = icon
	frame.description = description
	
	return frame
end

-- Showing options toggle button
function tABT:showOptionsButton()
	button = UI.CreateFrame("Texture", "TwinABTOptionsToggleButton", optionsContext)
	button:SetTexture("TwinABT", tABToptions.toggleButton.texture)
	button:SetWidth(tABToptions.toggleButton.width)
	button:SetHeight(tABToptions.toggleButton.height)
	button:SetVisible(true)
	if not tABToptions.toggleButton.x then
		tABToptions.toggleButton.x = (UIParent:GetWidth() - tABToptions.toggleButton.width) / 2
		tABToptions.toggleButton.y = (UIParent:GetHeight() - tABToptions.toggleButton.height) / 2
	end
	button:SetPoint("TOPLEFT", UIParent, "TOPLEFT", tABToptions.toggleButton.x, tABToptions.toggleButton.y)
	tABToptions.button = button
	
	-- Basic button states
	-- NORMAL
	function button.Event:MouseOut()
		self:SetTexture("TwinABT", tABToptions.toggleButton.texture)
	end
	-- HOVER
	function button.Event:MouseIn()
		self:SetTexture("TwinABT", tABToptions.toggleButton.textureHover)
	end
	-- FOCUS
	function button.Event:LeftDown()
		self:SetTexture("TwinABT", tABToptions.toggleButton.textureFocus)
	end
	-- FOCUS LOST
	function button.Event:LeftUp()
		self:SetTexture("TwinABT", tABToptions.toggleButton.texture)
	end
	
	-- Dragable
	function button.Event:RightDown()
		self.drag = true
	end
	function button.Event:RightUp()
		self.drag = false
		self.move = false
		self.offset = nil
		tABToptions.toggleButton.x = self:GetLeft()
		tABToptions.toggleButton.y = self:GetTop()
	end
	function button.Event:MouseMove(x,y)
		if self.drag then
			if not self.offset then
				self.offset = {
					x = x-self:GetLeft(),
					y = y-self:GetTop()
				}
			end
			local moveX, moveY = x-self.offset.x, y-self.offset.y
			button:SetPoint("TOPLEFT", UIParent, "TOPLEFT", moveX, moveY)
		end
	end
	
	-- Toggling Options Frame
	function button.Event:LeftClick()
		local display = not tABT.mainFrame:GetVisible()
		tABT.mainFrame:SetVisible(display)
		-- Buffs anchors
		tABTbuffs.EnableEditMode(display)
	end
end


-- Manages saved variables
function tABT:manageSavedVariables()
	if not tABToptions then tABToptions = tABToptionsDefaults end
	-- Options Window
	tABT:CreateOptionsFrame()
	tABT:CreateTabController()
	tABT:AddTabItems()
	--tABT:FillContent()
	-- Options toggle button
	tABT:showOptionsButton()
end


init()